# coding: utf-8

__doc__ = \
"""
*****************************************************************************

                      NAIADES QUESU CSV VALIDATOR
           Outil de contrôle des données au format QUESU CSV 3.1

Ce module regroupe les objets de gestion des erreurs de validation
"""
"\n*****************************************************************************\n\n"

import pathlib
from quesu_csv_type import *

class QuesuError():
    """Erreur d'interprétation d'une donnée QUESU CSV"""
    ERROR_LEVEL_FATAL = 'BLOQUANTE'
    ERROR_LEVEL_WARNING = 'NON_BLOQUANTE'
    ERROR_LEVEL_INFO = 'INFORMATION'

    ERROR_MSG_INVALID_FILE_NAME = "NOM DE FICHIER INCORRECT"
    ERROR_MSG_INVALID_ENCODING = "ENCODAGE INCORRECT"
    ERROR_MSG_INVALID_HEADER = "EN-TETE INCORRECT"
    ERROR_MSG_INVALID_CR = "RETOUR A LA LIGNE INCORRECT"
    ERROR_MSG_INVALID_RECORD_LENGTH = "NOMBRE DE CHAMPS INCORRECT"
    ERROR_MSG_NO_DATA = 'DONNNEE OBLIGATOIRE ABSENTE'
    ERROR_MSG_DATA_EMPTY = 'DONNNEE ABSENTE'
    ERROR_MSG_INVALID_FORMAT = 'FORMAT INCORRECT'
    ERROR_MSG_INVALID_REFERENCE = 'REFERENCE INVALIDE'

    def __init__(self, file, level, msg, line=0, col=0, field="", value="",content=""):
        """Constructeur

        file : nom du fichier QUESU CSV contenant la non conformité
        level : niveau de criticité de la non conformité
        msg : libellé de la non conformité
        line : n° de ligne dans le fichier QUESU CSV contenant la non conformité
        col : n° de colonne dans le fichier QUESU CSV contenant la non conformité
        field : champ QUESU CSV contenant la non conformité
        value : valeur de la non conformité

        """
        file = pathlib.Path(file)
        self.file = file.name
        self.level = level
        self.msg = msg
        self.line = line
        self.col = col
        self.field = field
        self.value = value
        self.content = content # contenu de la ligne en erreur

    def __str__(self):
        return ";".join([self.file, str(self.level), self.msg, str(self.line), str(
            self.col), self.field, self.value,self.content])

class QuesuErrorLogger():
    """Classe d'enregistrement des non conformités QUESU"""
    LOG_DIR_NAME = "quesu_error"
    LOG_EXT = ".csv"
    LOG_HEADER = "Fichier;Gravite;Erreur;Ligne;Colonne;Champ;Valeur"
    LOG_HEADER_ANA = LOG_HEADER + ";<CdStationMesureEauxSurface>;<CdPointEauxSurf>;<CodeSandreRdd>;<DateDebutOperationPrel>;<HeureDebutOperationPrel>;<DateFinOperationPrel>;<HeureFinOperationPrel>;<CdPrelevement>;<DatePrel>;<HeurePrel>;<DateFinPrel>;<HeureFinPrel>;<CoordXPrel>;<CoordYPrel>;<ProjectPrel>;<ZoneVerticaleProspectee>;<ProfondeurPrelevement>;<DifficultePrel>;<AccredPrel>;<FinalitePrel>;<AgrePrel>;<CommentairesPrel>;<CdSupport>;<CdMethodePrel>;<ProducteurPrel>;<Preleveur>;<RefAnaProd>;<DateAna>;<HeureAna>;<CdParametre>;<CdFractionAnalysee>;<RsAna>;<CdUniteMesure>;<RqAna>;<InsituAna>;<DifficulteAna>;<QualAna>;<CommentairesAna>;<ComResultatAna>;<StatutAna>;<AccreAna>;<LDAna>;<LQAna>;<LSAna>;<IncertAna>;<AgreAna>;<CdMethFractionnement>;<CdMethAna>;<RdtExtraction>;<CdMethExtraction>;<ProducteurAna>;<Laboratoire>;<FLG>"
    LOG_HEADER_CE = LOG_HEADER +  ";<CdStationMesureEauxSurface>;<CdPointEauxSurf>;<CdPrelevement>;<DatePrel>;<DateParEnv>;<HeureParEnv>;<CdParametre>;<RsParEnv>;<CdUniteMesure>;<RqParEn>;<QualParEnv>;<StatutParEn>;<ComParEnv>;<CdMethode>;<Producteur>;<Preleveur>;<FLG>"
    ERROR_PER_FILE = 1000000  # Nombre maximum d'erreurs par fichier
    ERROR_BUFFER_SIZE = 100000 # Nombre maximum d'erreurs en mémoire tampon

    def __init__(self, quesu_folder):
        """Initalise l'objet

        quesu_folder : objet pathlib.Path correspondant au répertoire de stockage des fichiers quesu contrôlés
        """
        self.quesu_log_path = None
        if quesu_folder.exists() :
            # Création du répertoire de conservation des logs
            quesu_log_path = quesu_folder.joinpath(__class__.LOG_DIR_NAME)
            if not quesu_log_path.exists():
                quesu_log_path.mkdir()

            self.quesu_log_path = quesu_log_path

        # Initialisation des statistiques d'analyse de chaque fichier CSV
        self.file_stats = {}

        # Initialisation des statistiques d'analyse du répertoire
        # error_count : nb d'erreur total
        # file_count : nb total de fichiers d'erreurs
        # temp_error_count : nb d'erreur temporaire (utile pour la gestion des fichiers)
        self.folder_stats = {
            QUESU_TYPE_PC_ANA:{
                QuesuError.ERROR_LEVEL_FATAL:{'error_count':0, 'file_count':0, 'temp_error_count':0},
                QuesuError.ERROR_LEVEL_WARNING:{'error_count':0, 'file_count':0, 'temp_error_count':0},
                QuesuError.ERROR_LEVEL_INFO:{'error_count':0, 'file_count':0, 'temp_error_count':0}
            },
            QUESU_TYPE_PC_CE:{
                QuesuError.ERROR_LEVEL_FATAL: {'error_count': 0, 'file_count': 0, 'temp_error_count':0},
                QuesuError.ERROR_LEVEL_WARNING: {'error_count': 0, 'file_count': 0, 'temp_error_count':0},
                QuesuError.ERROR_LEVEL_INFO: {'error_count': 0, 'file_count': 0, 'temp_error_count':0}
            },
            QUESU_TYPE_INCONNU:{
                QuesuError.ERROR_LEVEL_FATAL: {'error_count': 0, 'file_count': 0, 'temp_error_count':0},
                QuesuError.ERROR_LEVEL_WARNING: {'error_count': 0, 'file_count': 0, 'temp_error_count':0},
                QuesuError.ERROR_LEVEL_INFO: {'error_count': 0, 'file_count': 0, 'temp_error_count':0}
            }
        }

        # Liste des erreurs maintenues en mémoire pour limiter le nombre d'écritures sur disque
        self.errors = []

    def log(self, quesu_file_path, quesu_type, errors, force_write=True):
        """Enregistre une liste de non conformité

        quesu_file_path : chemin d'accès au fichier QUESU analysé
        quesu_type : type de données QUESU
        errors : liste d'objets QuesuError
        force_write : force l'écriture sur disque des erreurs en mémoire
        """
        if not self.quesu_log_path:
            return
        if not len(errors) and not force_write:
            return

        # Mémorise les erreurs pour limiter le nombre d'écritures sur disque
        self.errors.extend(errors)

        if len(self.errors) < __class__.ERROR_BUFFER_SIZE and not force_write:
            # On reporte l'écriture sur disque
            return

        # Détermination du fichier cible en fonction du type de données QUESU
        if quesu_type == QUESU_TYPE_PC_ANA:
            quesu_log_stem = "QUESU_CSV_ANA"
        elif quesu_type == QUESU_TYPE_PC_CE:
            quesu_log_stem = "QUESU_CSV_CEP"
        else:
            quesu_log_stem = "QUESU_CSV_INCONNU"

        # Création de liste de chaînes de caractères pour chaque niveau d'erreur
        if len(self.errors):
            errors_strings = {
                QuesuError.ERROR_LEVEL_FATAL:[str(error) + "\n" for error in self.errors if error.level == QuesuError.ERROR_LEVEL_FATAL],
                QuesuError.ERROR_LEVEL_WARNING:[str(error) + "\n" for error in self.errors if error.level == QuesuError.ERROR_LEVEL_WARNING],
                QuesuError.ERROR_LEVEL_INFO:[str(error) + "\n" for error in self.errors if error.level == QuesuError.ERROR_LEVEL_INFO]
            }
        else:
            errors_strings = {}

            # Enregistrement des non-conformités
        for error_level, errors_string in errors_strings.items():
            if not len(errors_string):
                continue

            log_file_path = self.quesu_log_path.joinpath(
                quesu_log_stem + "_err_" + error_level + "_" + str(
                    self.folder_stats[quesu_type][error_level]['file_count']) + __class__.LOG_EXT)
            if log_file_path.exists():
                mode = "a"
            else:
                mode = "w"
            with log_file_path.open(mode=mode, encoding="utf-8") as error_file:
                if mode == "w":
                    if quesu_type==QUESU_TYPE_PC_ANA :
                        error_file.write(__class__.LOG_HEADER_ANA + "\n")
                    elif quesu_type == QUESU_TYPE_PC_CE :     
                        error_file.write(__class__.LOG_HEADER_CE + "\n")
                    else :     
                        error_file.write(__class__.LOG_HEADER + "\n")
                error_file.writelines(errors_string)

            # Mise à jour des statistiques du fichier QUESU
            if not quesu_file_path in self.file_stats.keys():
                self.file_stats[quesu_file_path] = {QuesuError.ERROR_LEVEL_FATAL: 0,
                                                       QuesuError.ERROR_LEVEL_WARNING: 0,
                                                       QuesuError.ERROR_LEVEL_INFO: 0}
            self.file_stats[quesu_file_path][error_level] += len(errors_string)

            # Mise à jour des statistiques du répertoire
            self.folder_stats[quesu_type][error_level]['error_count'] += len(errors_string)
            self.folder_stats[quesu_type][error_level]['temp_error_count'] += len(errors_string)

            # Incrément du nb de fichiers de non-conformités si nécessaire
            if self.folder_stats[quesu_type][error_level]['temp_error_count'] / __class__.ERROR_PER_FILE >= 1:
                self.folder_stats[quesu_type][error_level]['temp_error_count'] = 0
                self.folder_stats[quesu_type][error_level]['file_count'] += 1

        # Réinitialisation du tampon de non-conformités
        self.errors = []
