# coding: utf-8
from profilehooks import profile

__doc__ = \
    """
    *****************************************************************************
    
                          NAIADES QUESU CSV VALIDATOR
               Outil de contrôle des données au format QUESU CSV 3.1
    
    Ce module définit la classe de validation d'une ligne de données QUESU CSV
    """
"\n*****************************************************************************\n\n"

from pyparsing import *
from quesu_csv_error import QuesuError

# Valeur par défaut pour l'unité de mesure PC (= sans objet)
PC_CdUniteMesure_FIELD_DEFAULT_VALUE = 'X'

# Paramétrage des analyses PC
PC_ANA_PARSE_INFO = (
    ('CdStationMesureEauxSurface', Word(alphanums + '-', max=8), QuesuError.ERROR_LEVEL_FATAL),
    ('CdPointEauxSurf', Word(nums, max=3) ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('CodeSandreRdd', Word(nums + '/', max=100) ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('DateDebutOperationPrel', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})'), QuesuError.ERROR_LEVEL_FATAL),
    ('HeureDebutOperationPrel', Regex(r'(\d{2}):(\d{2}):(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('DateFinOperationPrel', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('HeureFinOperationPrel', Regex(r'(\d{2}):(\d{2}):(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CdPrelevement', Word(printables + ' ', excludeChars=';', max=100), QuesuError.ERROR_LEVEL_FATAL),
    ('DatePrel', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})'), QuesuError.ERROR_LEVEL_FATAL),
    ('HeurePrel', Regex(r'(\d{2}):(\d{2}):(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('DateFinPrel', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('HeureFinPrel', Regex(r'(\d{2}):(\d{2}):(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CoordXPrel', Regex(r"[+-]?\d+[\,.]?\d*") ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CoordYPrel', Regex(r"[+-]?\d+[\,.]?\d*") ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('ProjectPrel', Word(nums, max=2) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('ZoneVerticaleProspectee', Word(nums, exact=1) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('ProfondeurPrelevement', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('DifficultePrel', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('AccredPrel', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('FinalitePrel', Word(nums, max=10), QuesuError.ERROR_LEVEL_FATAL),
    ('AgrePrel', oneOf("0 1 true false", ) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CommentairesPrel', Word(printables + alphas8bit + punc8bit + ' ’…', excludeChars=';') ^ Empty(), None),
    ('CdSupport', Word(nums, max=3), QuesuError.ERROR_LEVEL_FATAL),
    ('CdMethodePrel', Word(nums, max=5) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('ProducteurPrel', Word(alphanums, max=14), QuesuError.ERROR_LEVEL_WARNING),
    ('Preleveur', Word(alphanums, max=14) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('RefAnaProd', Word(printables + alphas8bit + ' ', excludeChars=';', max=100), QuesuError.ERROR_LEVEL_FATAL),
    ('DateAna', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('HeureAna', Regex(r'(\d{2}):(\d{2}):(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CdParametre', Word(nums, max=5), QuesuError.ERROR_LEVEL_FATAL),
    ('CdFractionAnalysee', Word(nums, max=3), QuesuError.ERROR_LEVEL_FATAL),
    # Pour permettre un contrôle métier impliquant les champs RsAna / RqAna, l'absence de valeur pour RsAna n'est pas
    # acceptée. Le contrôle est opéré lors du traitement de l'exception levée au moment du parsing
    ('RsAna', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?'), QuesuError.ERROR_LEVEL_FATAL),
    ('CdUniteMesure', Word(nums, max=5) ^ CaselessKeyword(PC_CdUniteMesure_FIELD_DEFAULT_VALUE), QuesuError.ERROR_LEVEL_FATAL),
    ('RqAna', Word(nums, max=2), QuesuError.ERROR_LEVEL_FATAL),
    ('InsituAna', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('DifficulteAna', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('QualAna', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('CommentairesAna', Word(printables + alphas8bit + punc8bit + ' ’…', excludeChars=';') ^ Empty(), None),
    ('ComResultatAna', Word(printables + alphas8bit + punc8bit + ' ’…', excludeChars=';') ^ Empty(), None),
    ('StatutAna', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('AccreAna', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('LDAna', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?') ^ Empty(), None),
    ('LQAna', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?') ^ Empty(), None),
    ('LSAna', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?') ^ Empty(), None),
    ('IncertAna', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?') ^ Empty(), None),
    ('AgreAna', oneOf("0 1 true false", ) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CdMethFractionnement', Word(nums, max=5) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CdMethAna', Word(nums, max=5), QuesuError.ERROR_LEVEL_FATAL),
    ('RdtExtraction', Regex(r'[+-]?(\d+([\,.]\d*)?|[\,.]\d+)([eE][+-]?\d+)?') ^ Empty(), None),
    ('CdMethExtraction', Word(nums, max=5) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('ProducteurAna', Word(alphanums, max=14), QuesuError.ERROR_LEVEL_WARNING),
    ('Laboratoire', Word(alphanums, max=14) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('FLG', CaselessKeyword('FLG'), QuesuError.ERROR_LEVEL_FATAL))

PC_ANA_CdStationMesureEauxSurface_FIELD_INDEX = 0
PC_ANA_CdParametre_FIELD_INDEX = 29
PC_ANA_CdUniteMesure_FIELD_INDEX = 32
PC_ANA_RqAna_FIELD_INDEX = 33

# Paramétrage des CEP
PC_CE_PARSE_INFO = (
    ('CdStationMesureEauxSurface', Word(alphanums + '-', max=8), QuesuError.ERROR_LEVEL_FATAL),
    ('CdPointEauxSurf', Word(nums, max=3) ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('CdPrelevement', Word(printables + ' ', excludeChars=';', max=100), QuesuError.ERROR_LEVEL_FATAL),
    ('DatePrel', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})'), QuesuError.ERROR_LEVEL_FATAL),
    ('DateParEnv', Regex(r'19(\d{2})-(\d{2})-(\d{2})') ^ Regex(r'20(\d{2})-(\d{2})-(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('HeureParEnv', Regex(r'(\d{2}):(\d{2}):(\d{2})') ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('CdParametre', Word(nums, max=5), QuesuError.ERROR_LEVEL_FATAL),
    # Pour permettre un contrôle métier impliquant les champs RsParEnv / RqParEn, l'absence de valeur pour RsAna n'est pas
    # acceptée. Le contrôle est opéré lors du traitement de l'exception levée au moment du parsing
    ('RsParEnv', Word(alphanums + '+-,.', max=15), QuesuError.ERROR_LEVEL_FATAL),
    ('CdUniteMesure', Word(nums, max=5) ^ CaselessKeyword(PC_CdUniteMesure_FIELD_DEFAULT_VALUE), QuesuError.ERROR_LEVEL_FATAL),
    ('RqParEn', Word(nums, max=2), QuesuError.ERROR_LEVEL_FATAL),
    ('QualParEnv', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('StatutParEn', Word(nums, exact=1), QuesuError.ERROR_LEVEL_FATAL),
    ('ComParEnv', Word(printables + alphas8bit + punc8bit + ' ’…', excludeChars=';') ^ Empty(), None),
    ('CdMethode', Word(nums, max=5) ^ Empty(), QuesuError.ERROR_LEVEL_WARNING),
    ('Producteur', Word(alphanums, max=14) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('Preleveur', Word(alphanums, max=14) ^ Empty(), QuesuError.ERROR_LEVEL_INFO),
    ('FLG', CaselessKeyword('FLG'), QuesuError.ERROR_LEVEL_FATAL))

PC_CEP_CdParametre_FIELD_INDEX = 6
PC_CEP_CdUniteMesure_FIELD_INDEX = 8
PC_CEP_RqParEnv_FIELD_INDEX = 9


class QuesuLine():
    """Classe de traitraitement d'une ligne de données QUESU"""
    QUESU_FIELD_SEPARATOR = ";"
    QUESU_LIST_SEPARATOR = "/"
    QUESU_EOF = "FLG"

    def __init__(self, quesu_type, quesu_parse_info, quesu_file_name="", quesu_empty_content=""):
        """Initialise l'objet

        quesu_type : type de données QUESU
        quesu_parse_info : paramétrage du parseur pour ce type de données
        quesu_file_name : fichier QUESU contenant les lignes à valider
        quesu_empty_content : ligne de données vide

        """
        self.quesu_type = quesu_type
        self.quesu_parse_info = quesu_parse_info
        self.quesu_file_name = quesu_file_name
        self.quesu_empty_content = quesu_empty_content

    def validate(self, quesu_line, quesu_references, quesu_line_num=0):
        """Valide la ligne de données QUESU

        quesu_line : ligne de données QUESU à valider
        quesu_references : dictionnaire des références SANDRE
        quesu_line_num : n° de la ligne dans le fichier QUESU à valider

        Renvoie un tableau des données interprétées et une liste des erreurs rencontrées
        """
        observation = []
        errors = []

        quesu_values = quesu_line.split(__class__.QUESU_FIELD_SEPARATOR)

        if len(quesu_values) != len(self.quesu_parse_info):
            # longueur d'enregistrement erronée
            errors.append(
                QuesuError(file=self.quesu_file_name, level=QuesuError.ERROR_LEVEL_FATAL,
                           msg=QuesuError.ERROR_MSG_INVALID_RECORD_LENGTH, line=quesu_line_num,
                           content=self.quesu_empty_content))
            return observation, errors

        quesu_field_index = 0
        quesu_col = 1
        found_error = False
        for quesu_value in quesu_values:
            quesu_field = self.quesu_parse_info[quesu_field_index][0]
            quesu_parse_expression = self.quesu_parse_info[quesu_field_index][1]
            quesu_error_level = self.quesu_parse_info[quesu_field_index][2]

            #
            # Validation syntaxique
            #
            try:
                quesu_parse_expression.parseString(quesu_value, parseAll=True)
            except ParseException:
                found_error = True
                if quesu_value == "":
                    if quesu_field == "RsAna":
                        # si RqAna est 0, 5 ou 6, on génère une anomalie non bloquante (anomalie bloquante sinon)
                        RqAna = quesu_values[PC_ANA_RqAna_FIELD_INDEX]
                        if RqAna != "0" and RqAna != "5" and RqAna != "6":
                            quesu_error_level = QuesuError.ERROR_LEVEL_FATAL
                        else:
                            found_error = False
                        quesu_err_msg = QuesuError.ERROR_MSG_DATA_EMPTY
                    elif quesu_field == "RsParEnv":
                        # si RqParAna est 0, 5 ou 6, on génère une anomalie non bloquante (anomalie bloquante sinon)
                        RqParEnv = quesu_values[PC_CEP_RqParEnv_FIELD_INDEX]
                        if RqParEnv != "0" and RqParEnv != "5" and RqParEnv != "6":
                            quesu_error_level = QuesuError.ERROR_LEVEL_FATAL
                        else:
                            found_error = False
                        quesu_err_msg = QuesuError.ERROR_MSG_DATA_EMPTY
                    else:
                        quesu_err_msg = QuesuError.ERROR_MSG_NO_DATA
                        quesu_error_level = QuesuError.ERROR_LEVEL_FATAL
                else:
                    quesu_err_msg = QuesuError.ERROR_MSG_INVALID_FORMAT

                if found_error:
                    errors.append(
                        QuesuError(file=self.quesu_file_name, level=quesu_error_level, msg=quesu_err_msg,
                                   line=quesu_line_num, col=quesu_col, field=quesu_field, value=quesu_value,
                                   content=quesu_line))

            #
            # Validation sémantique (existence des références SANDRE)
            #
            if not found_error and quesu_value != "":

                # Définition de la clef de la liste de valeurs possibles
                if quesu_field == "CdPointEauxSurf":
                    cdstation = quesu_values[PC_ANA_CdStationMesureEauxSurface_FIELD_INDEX]
                    key = "CdPointEauxSurf_" + cdstation
                elif quesu_field == "RsAna":
                    if quesu_values[PC_ANA_CdUniteMesure_FIELD_INDEX] == PC_CdUniteMesure_FIELD_DEFAULT_VALUE:
                        cdparametre = quesu_values[PC_ANA_CdParametre_FIELD_INDEX]
                        key = quesu_field + "_" + cdparametre
                    else:
                        key = quesu_field
                elif quesu_field == "RsParEnv":
                    if quesu_values[PC_CEP_CdUniteMesure_FIELD_INDEX] == PC_CdUniteMesure_FIELD_DEFAULT_VALUE:
                        cdparametre = quesu_values[PC_CEP_CdParametre_FIELD_INDEX]
                        key = quesu_field + "_" + cdparametre
                    else:
                        key = quesu_field
                else:
                    key = quesu_field

                if key in quesu_references.keys():
                    # Contrôle d'existence de la référence

                    # Traitement des valeurs de type LIST
                    if quesu_field == 'CodeSandreRdd':
                        quesu_refs = quesu_value.split(__class__.QUESU_LIST_SEPARATOR)
                    else:
                        quesu_refs = [quesu_value]

                    for quesu_ref in quesu_refs:
                        if quesu_ref not in quesu_references[key]:
                            found_error = True
                            quesu_error_msg = QuesuError.ERROR_MSG_INVALID_REFERENCE
                            errors.append(
                                QuesuError(file=self.quesu_file_name, level=quesu_error_level, msg=quesu_error_msg,
                                           line=quesu_line_num,
                                           col=quesu_col, field=quesu_field, value=quesu_ref, content=quesu_line))

            if found_error:
                observation.append((quesu_field, ""))
                found_error = False
            else:
                observation.append((quesu_field, quesu_value))
            quesu_field_index += 1
            quesu_col += len(quesu_value) + 1

        return observation, errors
