# coding: utf-8 

__version__ = "Version : 1.1_0"
__versionTime__ = "Dernière mise à jour : 4 Juillet 2017"
__author__ = "Auteur : AGENCE FRANCAISE POUR LA BIODIVERSITE\n" \
             "         Antonio Andrade <antonio.andrade@afbiodiversite.fr>"
__doc__ = \
"""
*****************************************************************************

                      NAIADES QUESU CSV VALIDATOR
        Outil de contrôle des données au format SANDRE QUESU CSV 3.1

"""
moduleinfo = __doc__ + "\n" + __version__ + "\n" + __versionTime__ + "\n" + __author__ + "\n" + \
"\n*****************************************************************************\n\n"

"""
*****************************************************************************

                      NAIADES QUESU CSV VALIDATOR
           Outil de contrôle des données au format QUESU CSV 3.1

Ce module constitue le point d'entrée de l'outil. Il traite la ligne de commande
et aiguille le traitement vers les différents modules spécifialisés

"\n*****************************************************************************\n\n"
"""

import pathlib
import os
import sys
from datetime import datetime
from quesu_csv_type import *
from quesu_csv_file import QuesuFile
from quesu_csv_reference import QuesuReference
from quesu_csv_error import QuesuErrorLogger

def showHelp():
    print("Usage : quesu_csv_validator.py -d <repertoire> [options]")
    print("\noptions:")
    print("   -h : affiche de l'aide")

if __name__ == '__main__':

    print(moduleinfo)

    if "-h" in sys.argv:
        showHelp()
        exit()

    if len(sys.argv) < 2 or not ("-d" in sys.argv):
        print("Erreur : ligne de commande incorrecte")
        showHelp()
        exit()

    csvFolder = pathlib.Path(sys.argv[sys.argv.index("-d") + 1])
    if not csvFolder.exists() or not csvFolder.is_dir():
        print("Répertoire {0} introuvable. Veuillez spécifier un nom de répertoire valide.".format(csvFolder))
        exit()

    start = datetime.now()
    print("{0} : contrôle du répertoire {1}\n".format(start.strftime("%d/%m/%y %H:%M:%S"), csvFolder))

    # Initialisation des statistiques de validation
    # Pour chaque type de données, l'outil comptabilise :
    # - le nombre de fichiers traités
    # - le nombre d'observations traitées
    # - le nombre total d'erreurs identifiées
    # - le nombre d'erreurs classées "fatale"
    # - le nombre d'erreurs classées "alerte"
    # - le nombre d'erreurs classées "information"
    validation_stats = {
        QUESU_TYPE_PC_ANA:[0, 0, 0, 0, 0, 0],
        QUESU_TYPE_PC_CE: [0, 0, 0, 0, 0, 0],
        QUESU_TYPE_INCONNU: [0, 0, 0, 0, 0, 0]
    }

    # Chargement des référentiels du Sandre
    print("{0} : chargement des référentiels du SANDRE".format(datetime.now().strftime("%d/%m/%y %H:%M:%S")))
    quesu_references = QuesuReference(csvFolder).getReferences()
    if quesu_references is None:
        print("{0} : erreur de chargement des référentiels du SANDRE".format(datetime.now().strftime("%d/%m/%y %H:%M:%S")))
        exit(1)

    # Initialisation de l'objet d'enregistrement des non conformités QUESU
    quesu_logger = QuesuErrorLogger(quesu_folder=csvFolder)

    for csvFile in list(csvFolder.glob('./*.csv')):

        startFileProcessing = datetime.now()
        print("{0} : contrôle du fichier {1}".format(startFileProcessing.strftime("%d/%m/%y %H:%M:%S"), os.path.basename(str(csvFile))))

        file_validator = QuesuFile(csvFile, quesu_logger)
        results = file_validator.validate(quesu_references=quesu_references)

        validation_stats[file_validator.quesu_type][0] += 1
        validation_stats[file_validator.quesu_type][1] += results['line_count']
        validation_stats[file_validator.quesu_type][2] += results['error_count']
        validation_stats[file_validator.quesu_type][3] += results['fatal_count']
        validation_stats[file_validator.quesu_type][4] += results['warning_count']
        validation_stats[file_validator.quesu_type][5] += results['info_count']

        stopFileProcessing = datetime.now()
        print("{0} : => durée du contrôle ({1})".format(stopFileProcessing.strftime("%d/%m/%y %H:%M:%S"),
                                                        stopFileProcessing - startFileProcessing))
        print("{0} : => {1} observations contrôlées".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"), results['line_count']))
        print("{0} : => {1} erreurs détectées : {2} fatales / {3} alertes / {4} informations".format(
            datetime.now().strftime("%d/%m/%y %H:%M:%S"), results['error_count'], results['fatal_count'],
            results['warning_count'], results['info_count']))

    stop = datetime.now()
    print("\n{0} : => durée totale du contrôle ({1})".format(stop.strftime("%d/%m/%y %H:%M:%S"), stop - start))
    print("{0} : => {1} fichiers d'analyses PC contrôlés".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"), validation_stats[QUESU_TYPE_PC_ANA][0]))
    print("{0} : => {1} analyses PC contrôlées".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"),
                                                                 validation_stats[QUESU_TYPE_PC_ANA][1]))
    print("{0} : => {1} erreurs détectées : {2} bloquantes / {3} non bloquantes / {4} informations".format(
        datetime.now().strftime("%d/%m/%y %H:%M:%S"), validation_stats[QUESU_TYPE_PC_ANA][2], validation_stats[QUESU_TYPE_PC_ANA][3],
        validation_stats[QUESU_TYPE_PC_ANA][4], validation_stats[QUESU_TYPE_PC_ANA][5]))
    print("{0} : => {1} fichiers de CEP contrôlés".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"),
                                                                 validation_stats[QUESU_TYPE_PC_CE][0]))
    print("{0} : => {1} CEP contrôlées".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"),
                                                       validation_stats[QUESU_TYPE_PC_CE][1]))
    print("{0} : => {1} erreurs détectées : {2} bloquantes / {3} non bloquantes / {4} informations".format(
        datetime.now().strftime("%d/%m/%y %H:%M:%S"), validation_stats[QUESU_TYPE_PC_CE][2],
        validation_stats[QUESU_TYPE_PC_CE][3],
        validation_stats[QUESU_TYPE_PC_CE][4], validation_stats[QUESU_TYPE_PC_CE][5]))
    print("{0} : => {1} fichiers de type INCONNU contrôlés".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"),
                                                          validation_stats[QUESU_TYPE_INCONNU][0]))
    print("{0} : => {1} CEP contrôlées".format(datetime.now().strftime("%d/%m/%y %H:%M:%S"),
                                               validation_stats[QUESU_TYPE_INCONNU][1]))
    print("{0} : => {1} erreurs détectées : {2} bloquantes / {3} non bloquantes / {4} informations".format(
        datetime.now().strftime("%d/%m/%y %H:%M:%S"), validation_stats[QUESU_TYPE_INCONNU][2],
        validation_stats[QUESU_TYPE_INCONNU][3],
        validation_stats[QUESU_TYPE_INCONNU][4], validation_stats[QUESU_TYPE_INCONNU][5]))

    print("\nVoir le détail des erreurs dans les fichiers *_err*.csv du sous répertoire quesu_error")
